# -*- coding: utf-8 -*-
"""
    Appearance Config
    ~~~~~~~~~~~~~~~~~
    Configuration for an annotation's appearance.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import attr

from pdf_annotate.config.constants import ALLOWED_ALIGNS
from pdf_annotate.config.constants import ALLOWED_BASELINES
from pdf_annotate.config.constants import ALLOWED_LINE_CAPS
from pdf_annotate.config.constants import ALLOWED_LINE_JOINS
from pdf_annotate.config.constants import BLACK
from pdf_annotate.config.constants import DEFAULT_BORDER_STYLE
from pdf_annotate.config.constants import DEFAULT_CONTENT
from pdf_annotate.config.constants import DEFAULT_FONT_SIZE
from pdf_annotate.config.constants import DEFAULT_LINE_SPACING
from pdf_annotate.config.constants import DEFAULT_STROKE_WIDTH
from pdf_annotate.config.constants import GRAPHICS_STATE_NAME
from pdf_annotate.config.constants import TEXT_ALIGN_LEFT
from pdf_annotate.config.constants import TEXT_BASELINE_MIDDLE
from pdf_annotate.config.graphics_state import GraphicsState
from pdf_annotate.graphics import ContentStream
from pdf_annotate.graphics import FillColor
from pdf_annotate.graphics import GraphicsState as CSGraphicsState
from pdf_annotate.graphics import Stroke
from pdf_annotate.graphics import StrokeAndFill
from pdf_annotate.graphics import StrokeColor
from pdf_annotate.graphics import StrokeWidth
from pdf_annotate.util.validation import between
from pdf_annotate.util.validation import Boolean
from pdf_annotate.util.validation import Color
from pdf_annotate.util.validation import Enum
from pdf_annotate.util.validation import Field
from pdf_annotate.util.validation import Number
from pdf_annotate.util.validation import positive
from pdf_annotate.util.validation import String
from pdf_annotate.util.validation import validate_dash_array


def is_transparent(color):
    # E.g. a soothing gray: [0, 0, 0, 0.5]
    if color is None:
        return False
    return len(color) == 4 and color[-1] < 1


@attr.s
class Appearance(object):
    # Stroke attributes
    stroke_color = Color(default=BLACK)
    stroke_width = Number(default=DEFAULT_STROKE_WIDTH, validator=positive)
    border_style = String(default=DEFAULT_BORDER_STYLE)
    dash_array = Field(list, default=None, validator=validate_dash_array)
    line_cap = Enum(ALLOWED_LINE_CAPS, default=None)
    line_join = Enum(ALLOWED_LINE_JOINS, default=None)
    miter_limit = Number(default=None, validator=positive)
    stroke_transparency = Number(default=None, validator=between(0, 1))

    # Fill attributes
    fill = Color(default=None)
    fill_transparency = Number(default=None, validator=between(0, 1))

    # Text attributes
    content = String(default=DEFAULT_CONTENT)
    font_size = Number(default=DEFAULT_FONT_SIZE, validator=positive)
    text_align = Enum(ALLOWED_ALIGNS, default=TEXT_ALIGN_LEFT)
    text_baseline = Enum(ALLOWED_BASELINES, default=TEXT_BASELINE_MIDDLE)
    line_spacing = Number(default=DEFAULT_LINE_SPACING, validator=positive)
    wrap_text = Boolean(default=True)

    # Image attributes
    image = String(default=None)

    # Advanced attributes
    appearance_stream = Field(ContentStream, default=None)
    xobjects = Field(dict, default=None)
    graphics_states = Field(dict, default=None)
    fonts = Field(dict, default=None)

    def copy(self, **kwargs):
        A = Appearance(**kwargs)
        for k, v in self.__dict__.items():
            if k not in kwargs:
                setattr(A, k, v)
        return A

    def _get_stroke_transparency(self):
        stroke_transparency = None
        if is_transparent(self.stroke_color):
            stroke_transparency = self.stroke_color[-1]
        if self.stroke_transparency is not None:
            stroke_transparency = self.stroke_transparency
        return stroke_transparency

    def _get_fill_transparency(self):
        fill_transparency = None
        if is_transparent(self.fill):
            fill_transparency = self.fill[-1]
        if self.fill_transparency is not None:
            fill_transparency = self.fill_transparency
        return fill_transparency

    def get_graphics_state(self):
        """Return a GraphicsState config from the appearance's graphics-state-
        applicable params.

        :returns GraphicsState:
        """
        return GraphicsState(
            dash_array=self.dash_array,
            line_cap=self.line_cap,
            line_join=self.line_join,
            miter_limit=self.miter_limit,
            stroke_transparency=self._get_stroke_transparency(),
            fill_transparency=self._get_fill_transparency(),
        )


def set_appearance_state(stream, A):
    """Update the graphics command stream to reflect appearance properties.

    :param ContentStream stream: current content stream
    :param Appearance A: appearance object
    """
    # Add in the `gs` command, which will execute the named graphics state from
    # the Resources dict, and set CA and/or ca values. The annotations
    # themselves will need to ensure that the proper ExtGState object is
    # present in the Resources dict.
    graphics_state = A.get_graphics_state()
    if graphics_state.has_content():
        stream.add(CSGraphicsState(GRAPHICS_STATE_NAME))

    stream.extend([
        StrokeColor(*A.stroke_color[:3]),
        StrokeWidth(A.stroke_width),
    ])

    # TODO support more color spaces - CMYK and GrayScale
    if A.fill is not None:
        stream.add(FillColor(*A.fill[:3]))


def stroke_or_fill(stream, A):
    if A.fill is not None:
        stream.add(StrokeAndFill())
    else:
        stream.add(Stroke())



# -*- coding: utf-8 -*-
"""
    Text Annotations
    ~~~~~~~~~~~~~~~~
    The FreeText annotation

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import os.path

from pdfrw import IndirectPdfDict
from pdfrw import PdfArray
from pdfrw import PdfDict
from pdfrw import PdfName
from pdfrw import PdfString

from pdf_annotate.annotations.base import _make_border_dict
from pdf_annotate.annotations.base import Annotation
from pdf_annotate.config.constants import DEFAULT_BASE_FONT
from pdf_annotate.config.constants import GRAPHICS_STATE_NAME
from pdf_annotate.config.constants import PDF_ANNOTATOR_FONT
from pdf_annotate.graphics import BeginText
from pdf_annotate.graphics import ContentStream
from pdf_annotate.graphics import EndText
from pdf_annotate.graphics import FillColor
from pdf_annotate.graphics import Font
from pdf_annotate.graphics import GraphicsState as CSGraphicsState
from pdf_annotate.graphics import Restore
from pdf_annotate.graphics import Save
from pdf_annotate.graphics import Text
from pdf_annotate.graphics import TextMatrix
from pdf_annotate.util.geometry import translate
from pdf_annotate.util.text import get_wrapped_lines
from pdf_annotate.util.true_type_font import get_true_type_font


HELVETICA_PATH = os.path.join(
    os.path.dirname(__file__),
    '..',
    'fonts',
    'Helvetica.ttf',
)


class FreeText(Annotation):
    """FreeText annotation. Right now, we only support writing text in the
    Helvetica font. Dealing with fonts is tricky business, so we'll leave that
    for later.
    """
    subtype = 'FreeText'

    def make_rect(self):
        L = self._location
        return [L.x1, L.y1, L.x2, L.y2]

    def make_default_appearance(self):
        """Returns a DA string for the text object, e.g. '1 0 0 rg /Helv 12 Tf'
        """
        A = self._appearance
        stream = ContentStream([
            FillColor(*A.fill[:3]),
            Font(PDF_ANNOTATOR_FONT, A.font_size),
        ])
        return stream.resolve()

    def add_additional_pdf_object_data(self, obj):
        obj.Contents = self._appearance.content
        obj.DA = self.make_default_appearance()
        obj.C = []
        # TODO allow setting border on free text boxes
        obj.BS = _make_border_dict(width=0, style='S')
        # TODO DS is required to have BB not redraw the annotation in their own
        # style when you edit it.

    @staticmethod
    def make_font_file_object(tt_font):
        """Make an embedded font object from the true type font itself.

        :param TrueTypeFont tt_font: Our utility class used to parse and calculate font metrics
        from a true type font.
        :returns PdfDict: font file PdfDict object stream.
        """
        # TODO: make subset font here
        with open(tt_font.ttfPath, 'rb') as font_file:
            data = font_file.read()

        # Let's let pdfrw handle compressing streams
        return IndirectPdfDict(stream=data.decode('Latin-1'))

    @staticmethod
    def make_to_unicode_object():
        """Make a toUnicode object which allows the PDF reader to derive content from the PDF
        with the CIDFont embedded.  This map converts from CIDs to Unicode code points.

        :returns PdfDict: toUnicode CMAP PdfDict object.
        """
        # See section 9.10.3 ToUnicode CMaps of PDF 1.6 Spec
        # TODO: For now we put an empty mapping in.
        return IndirectPdfDict(stream='\n'.join((
            "/CIDInit /ProcSet findresource begin",
            "12 dict begin",
            "begincmap",
            "/CIDSystemInfo",
            "<</Registry (Adobe)",
            "/Ordering (UCS)",
            "/Supplement 0",
            ">> def",
            "/CMapName /Adobe-Identity-UCS def",
            "/CMapType 2 def",
            "1 begincodespacerange",
            "<0000> <FFFF>",
            "endcodespacerange",
            "1 beginbfrange",
            "<0000> <FFFF> <0000>",
            "endbfrange",
            "endcmap",
            "CMapName currentdict /CMap defineresource pop",
            "end",
            "end"
        )))

    @staticmethod
    def make_cid_to_gid_map_object(tt_font):
        """Make a CID to GID map that is used to map character ids to glyph ids in the font.

        :param TrueTypeFont tt_font: Our utility class used to parse and calculate font metrics
        from a true type font.
        :returns PdfDict: CIDtoGID PdfDict object.
        """
        # Let's make this as large as possibly addressable for now, it will compress nicely.
        mapping_size = 256 * 256
        cid_to_gid_map = ["\x00"] * mapping_size * 2

        for cc, glyph_name in tt_font.metrics.cmap.items():
            # TODO: What is the expectation here since PDF only supports two bytes lookups?
            if cc >= mapping_size:
                continue
            glyph_id = tt_font.get_glyph_id(glyph_name)
            cid_to_gid_map[cc * 2] = chr(glyph_id >> 8)
            cid_to_gid_map[cc * 2 + 1] = chr(glyph_id & 0xFF)
        cid_to_gid_map = ''.join(cid_to_gid_map)

        # Let's let pdfrw handle the compressing of streams
        return IndirectPdfDict(stream=cid_to_gid_map)

    @staticmethod
    def make_font_descriptor_object(tt_font):
        """Make a Font Descriptor object containing some calculated metrics
        for the font.

        :param TrueTypeFont tt_font: Our utility class used to parse and calculate font metrics
        from a true type font.
        :returns PdfDict: Font Descriptor PdfDict object.
        """
        return IndirectPdfDict(
            Type=PdfName('FontDescriptor'),
            FontName=PdfName(tt_font.fontName),
            Flags=tt_font.metrics.flags,
            FontBBox=tt_font.metrics.bbox,
            ItalicAngle=int(tt_font.metrics.italicAngle),
            Ascent=int(round(tt_font.metrics.ascent, 0)),
            Descent=int(round(tt_font.metrics.descent, 0)),
            CapHeight=int(round(tt_font.metrics.capHeight, 0)),
            StemV=int(round(tt_font.metrics.stemV, 0)),
            MissingWidth=int(round(tt_font.metrics.defaultWidth, 0)),
            FontFile2=FreeText.make_font_file_object(tt_font)
        )

    @staticmethod
    def make_cid_system_info_object():
        """Make a CID System Info object.

        :returns PdfDict: CID System Info PdfDict object.
        """
        return IndirectPdfDict(
            Registry=PdfString('(Adobe)'),
            Ordering=PdfString('(UCS)'),
            Supplement=0
        )

    @staticmethod
    def make_cid_font_object(tt_font):
        """Make a CID Type 2 font object for including as a descendant of a composite
        Type 0 font object.

        :param TrueTypeFont tt_font: Our utility class used to parse and calculate font metrics
        from a true type font.
        :returns PdfDict: CID Font Type 2 PdfDict object.
        """
        return IndirectPdfDict(
            Type=PdfName('Font'),
            Subtype=PdfName('CIDFontType2'),
            BaseFont=PdfName(tt_font.fontName),
            CIDSystemInfo=FreeText.make_cid_system_info_object(),
            FontDescriptor=FreeText.make_font_descriptor_object(tt_font),
            DW=int(round(tt_font.metrics.defaultWidth, 0)),
            Widths=PdfArray(tt_font.metrics.widths),
            CIDToGIDMap=FreeText.make_cid_to_gid_map_object(tt_font),
        )

    @staticmethod
    def make_composite_font_object(font_file_path):
        """Make a PDF Type0 composite font object for embedding in the annotation's
        Resources dict.

        :param str font_file_path: The path and filename to the true type font we want to embed.
        :returns PdfDict: Resources PdfDict object, ready to be included in the
            Resources 'Font' subdictionary.
        """
        # TODO: Get font name from font program itself
        tt_font = get_true_type_font(font_file_path, DEFAULT_BASE_FONT)

        return IndirectPdfDict(
            Type=PdfName('Font'),
            Subtype=PdfName('Type0'),
            BaseFont=PdfName(tt_font.fontName),
            Encoding=PdfName('Identity-H'),
            DescendantFonts=PdfArray([
                FreeText.make_cid_font_object(tt_font)
            ]),
            ToUnicode=FreeText.make_to_unicode_object()
        )

    @staticmethod
    def make_font_object():
        """Make a PDF Type1 font object for embedding in the annotation's
        Resources dict. Only Helvetica is supported as a base font.

        :returns PdfDict: Resources PdfDict object, ready to be included in the
            Resources 'Font' subdictionary.
        """
        return PdfDict(
            Type=PdfName('Font'),
            Subtype=PdfName('Type1'),
            BaseFont=PdfName(DEFAULT_BASE_FONT),
            Encoding=PdfName('WinAnsiEncoding'),
        )

    def add_additional_resources(self, resources):
        font_dict = PdfDict()
        font_dict[PdfName(PDF_ANNOTATOR_FONT)] = self.make_font_object()
        resources[PdfName('Font')] = font_dict

    def make_appearance_stream(self):
        A = self._appearance
        L = self._location

        stream = ContentStream([
            Save(),
            BeginText(),
            FillColor(*A.fill[:3]),
            Font(PDF_ANNOTATOR_FONT, A.font_size),
        ])

        graphics_state = A.get_graphics_state()
        if graphics_state.has_content():
            stream.add(CSGraphicsState(GRAPHICS_STATE_NAME))

        # Actually draw the text inside the rectangle
        stream.extend(get_text_commands(
            L.x1, L.y1, L.x2, L.y2,
            text=A.content,
            font_size=A.font_size,
            wrap_text=A.wrap_text,
            align=A.text_align,
            baseline=A.text_baseline,
            line_spacing=A.line_spacing,
        ))
        stream.extend([
            EndText(),
            Restore(),
        ])

        return stream


def get_text_commands(
    x1, y1, x2, y2,
    text, font_size, wrap_text,
    align, baseline, line_spacing,
):
    """Return the graphics stream commands necessary to render a free text
    annotation, given the various parameters.

    Text is optionally wrapped, then arranged according to align (horizontal
    alignment), and baseline (vertical alignment).

    :param number x1: bounding box lower left x
    :param number y1: bounding box lower left y
    :param number x2: bounding box upper right x
    :param number y2: bounding box upper right y
    :param str text: text to add to annotation
    :param number font_size: font size
    :param bool wrap_text: whether to wrap the text
    :param str align: 'left'|'center'|'right'
    :param str baseline: 'top'|'middle'|'bottom'
    :param number line_spacing: multiplier to determine line spacing
    """
    font = get_true_type_font(
        path=HELVETICA_PATH,
        font_name=DEFAULT_BASE_FONT,
        font_size=font_size,
    )

    lines = get_wrapped_lines(
        text=text,
        measure=font.measure_text,
        max_length=x2 - x1,
    ) if wrap_text else [text]
    # Line breaking cares about the whitespace in the string, but for the
    # purposes of laying out the broken lines, we want to measure the lines
    # without trailing/leading whitespace.
    lines = [line.strip() for line in lines]
    y_coords = _get_vertical_coordinates(
        lines,
        y1,
        y2,
        font_size,
        line_spacing,
        baseline,
    )
    xs = _get_horizontal_coordinates(lines, x1, x2, font.measure_text, align)
    commands = []
    for line, x, y in zip(lines, xs, y_coords):
        commands.extend([
            TextMatrix(translate(x, y)),
            Text(line),
        ])
    return commands


def _get_vertical_coordinates(
    lines,
    y1,
    y2,
    font_size,
    line_spacing,
    baseline,
):
    """Calculate vertical coordinates for all the lines at once, honoring the
    text baseline property.
    """
    line_spacing = font_size * line_spacing
    if baseline == 'top':
        first_y = y2 - line_spacing
    elif baseline == 'middle':
        midpoint = (y2 + y1) / 2.0
        # For the first line of vertically centered text, go up half the # of
        # lines, then go back down half the font size.
        first_y = midpoint - \
            (line_spacing - font_size) + \
            (((len(lines) - 1) / 2.0) * line_spacing)
    else:  # bottom
        first_y = y1 + \
            (line_spacing - font_size) + \
            (line_spacing * (len(lines) - 1))
    return [first_y - (i * line_spacing) for i in range(len(lines))]


def _get_horizontal_coordinates(lines, x1, x2, measure, align):
    # NOTE: this padding is to keep text annotations as they are from cutting
    # off text at the edges in certain conditions. The annotation rectangle
    # and how PDFs draw text needs to be revisited, as this padding shouldn't
    # be necessary.
    PADDING = 1
    if align == 'left':
        return [x1 + PADDING for _ in range(len(lines))]
    elif align == 'center':
        widths = [measure(line) for line in lines]
        max_width = x2 - x1
        return [x1 + ((max_width - width) / 2.0) - PADDING for width in widths]
    else:  # right
        widths = [measure(line) for line in lines]
        max_width = x2 - x1
        return [x1 + (max_width - width) - PADDING for width in widths]



# -*- coding: utf-8 -*-
"""
    PdfAnnotator
    ~~~~~~~~~~~~
    The core annotator class.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import warnings

from pdfrw import PdfReader
from pdfrw import PdfWriter

from pdf_annotate.annotations.image import Image
from pdf_annotate.annotations.points import Ink
from pdf_annotate.annotations.points import Line
from pdf_annotate.annotations.points import Polygon
from pdf_annotate.annotations.points import Polyline
from pdf_annotate.annotations.rect import Circle
from pdf_annotate.annotations.rect import Square
from pdf_annotate.annotations.text import FreeText
from pdf_annotate.config.metadata import Metadata
from pdf_annotate.config.metadata import UNSET
from pdf_annotate.graphics import ContentStream
from pdf_annotate.util.geometry import identity
from pdf_annotate.util.geometry import matrix_multiply
from pdf_annotate.util.geometry import normalize_rotation
from pdf_annotate.util.geometry import rotate
from pdf_annotate.util.geometry import scale
from pdf_annotate.util.geometry import translate
from pdf_annotate.util.validation import NUMERIC_TYPES


NAME_TO_ANNOTATION = {
    'square': Square,
    'circle': Circle,
    'line': Line,
    'polygon': Polygon,
    'polyline': Polyline,
    'ink': Ink,
    'text': FreeText,
    'image': Image,
}


class PDF(object):

    def __init__(self, pdf_reader):
        self._reader = pdf_reader
        self.pdf_version = self._reader.private.pdfdict.version

    def get_page(self, page_number):
        if page_number > len(self._reader.pages) - 1:
            raise ValueError('Page number {} out of bounds ({} pages)'.format(
                page_number,
                len(self._reader.pages),
            ))
        return self._reader.pages[page_number]

    def get_rotation(self, page_number):
        """Returns the rotation of a specified page."""
        page = self.get_page(page_number)
        rotate = int(page.inheritable.Rotate or 0)
        return normalize_rotation(rotate)


class PdfAnnotator(object):

    def __init__(self, file_or_reader, scale=None, compress=True):
        """Draw annotations directly on PDFs. Annotations are always drawn on
        as if you're drawing them in a viewer, i.e. they take into account page
        rotation and weird, translated coordinate spaces.

        :param str|PdfReader file_or_reader: filename of PDF or pdfrw.PdfReader
        :param number|tuple|None scale: number by which to scale coordinates
            to get to default user space. Use this if, for example, your points
            in the coordinate space of the PDF viewed at a dpi. In this case,
            scale would be 72/dpi. Can also specify a 2-tuple of x and y scale.
        :param bool compress: whether to output flate-compressed PDFs
        """
        if isinstance(file_or_reader, str):
            file_or_reader = PdfReader(file_or_reader)
        self._pdf = PDF(file_or_reader)
        self._scale = self._expand_scale(scale)
        self._dimensions = {}
        self._compress = compress

    def _expand_scale(self, scale):
        if scale is None:
            return 1, 1
        elif isinstance(scale, NUMERIC_TYPES):
            return (scale, scale)
        return scale

    def set_page_dimensions(self, dimensions, page_number):
        """Set dimensions for a given page number. If set, the dimensions for
        this page override the document-wide rotation and scale settings.

        :param tuple|None dimensions: As a convenient alternative to scale and
            you can pass in the dimensions of your sheet when viewed in a
            certain setting. For example, an 8.5"x11" PDF, rotated at 90° and
            rastered at 150 dpi, would produce dimensions of (1650, 1275). If
            you pass this in, you can then specify your coordinates in this
            coordinate space.
        :param int page_number:
        """
        self._dimensions[page_number] = dimensions

    def get_page_bounding_box(self, page_number):
        page = self._pdf.get_page(page_number)
        # PDF bounding boxes are complicated. We choose to use the CropBox, if
        # it's available, because that's what Acrobat uses to display the
        # actual PDF, and what pdfinfo uses to determine the PDF's dimensions.
        # If CropBox isn't available, we use MediaBox. We ignore the TrimBox
        # because Acrobat also ignores this when displaying the PDF.
        crop_box = page.inheritable.CropBox
        if crop_box is not None:
            return [float(n) for n in crop_box]
        return [float(n) for n in page.inheritable.MediaBox]

    def get_size(self, page_number):
        """Returns the size of the specified page's bounding box (pts),
        accounting for page rotation.

        :param int page_number:
        :returns tuple: If page is rotated 90° or 270°, the returned value will
            be (height, width) in PDF user space. Otherwise the returned value
            will be (width, height).
        """
        x1, y1, x2, y2 = self.get_page_bounding_box(page_number)
        rotation = self._pdf.get_rotation(page_number)

        if rotation in (0, 180):
            return (x2 - x1, y2 - y1)

        return (y2 - y1, x2 - x1)

    def add_annotation(
        self,
        annotation_type,
        location,
        appearance,
        metadata=None,
    ):
        """Add an annotation of the given type, with the given parameters, to
        the given location of the PDF.

        :param str annotation_type: E.g. 'square'
        :param Location location: Annotation's Location object, specified in
            the coordinate system of the client. Coordinates will be
            transformed to PDF user space via get_transform.
        :param Appearance appearance:
        :param Metadata|None|UNSET metadata: Metadata object. If UNSET, no
            metadata is written on the entire annotation. If None, default
            metadata is used.
        """
        self._before_add(location)
        metadata = self._resolve_metadata(metadata)
        self._validate_appearance_stream(appearance)
        annotation = self.get_annotation(
            annotation_type,
            location,
            appearance,
            metadata,
        )
        self._add_annotation(annotation)

    @staticmethod
    def _resolve_metadata(metadata):
        if isinstance(metadata, Metadata):
            return metadata
        elif metadata is None:
            return Metadata()
        elif metadata is UNSET:
            return None
        else:
            raise ValueError('Invalid metadata')

    @staticmethod
    def _validate_appearance_stream(appearance):
        stream = appearance.appearance_stream
        if stream is not None and not isinstance(stream, ContentStream):
            raise ValueError(
                'Invalid appearance stream format: {}'.format(type(stream)))

    def _before_add(self, location):
        # Steps to take before trying to add an annotation to `location`
        page = self._pdf.get_page(location.page)
        user_unit = page.inheritable.UserUnit
        if user_unit not in (1, None):
            warnings.warn(
                'Unsupported UserUnit (value: {})'.format(user_unit)
            )

    def get_annotation(self, annotation_type, location, appearance, metadata):
        # TODO filter on valid PDF versions, by type
        annotation_cls = NAME_TO_ANNOTATION.get(annotation_type)
        if annotation_cls is None:
            raise ValueError('Invalid/unsupported annotation type: {}'.format(
                annotation_type
            ))

        annotation = annotation_cls(location, appearance, metadata)
        annotation.validate(self._pdf.pdf_version)
        return annotation

    def get_scale(self, page_number):
        """Public API to get the x and y scales of the given page.

        :param int page_number:
        :returns 2-tuple: (x_scale, y_scale)
        """
        rotation = self._pdf.get_rotation(page_number)
        bounding_box = self.get_page_bounding_box(page_number)
        return self._get_scale(page_number, bounding_box, rotation)

    def get_rotation(self, page_number):
        """Public API to get the rotation of the give page.

        :param int page_number:
        :returns int: integer where i % 90 == 0
        """
        return self._pdf.get_rotation(page_number)

    def _get_scale(self, page_number, bounding_box, rotation):
        W = bounding_box[2] - bounding_box[0]
        H = bounding_box[3] - bounding_box[1]

        dimensions = self._dimensions.get(page_number)
        if dimensions is not None:
            # User-specified dimensions for a particular page just give us the
            # scaling factor to use for that page.
            width_d, height_d = dimensions
            width_pts, height_pts = W, H
            if rotation in (90, 270):
                width_pts, height_pts = H, W
            x_scale = (width_pts / float(width_d))
            y_scale = (height_pts / float(height_d))
        else:
            x_scale, y_scale = self._scale

        return x_scale, y_scale

    def get_transform(self, page_number, rotation):
        bounding_box = self.get_page_bounding_box(page_number)
        _scale = self._get_scale(page_number, bounding_box, rotation)
        return self._get_transform(bounding_box, rotation, _scale)

    @staticmethod
    def _get_transform(bounding_box, rotation, _scale):
        """Get the transformation required to go from the user's desired
        coordinate space to PDF user space, taking into account rotation,
        scaling, translation (for things like weird media boxes).
        """
        # Unrotated width and height, in pts
        W = bounding_box[2] - bounding_box[0]
        H = bounding_box[3] - bounding_box[1]

        scale_matrix = scale(*_scale)

        x_translate = 0 + bounding_box[0]
        y_translate = 0 + bounding_box[1]
        mb_translate = translate(x_translate, y_translate)

        # Because of how rotation works the point isn't rotated around an axis,
        # but the axis itself shifts. So we have to represent the rotation as
        # rotation + translation.
        rotation_matrix = rotate(rotation)

        translate_matrix = identity()
        if rotation == 90:
            translate_matrix = translate(W, 0)
        elif rotation == 180:
            translate_matrix = translate(W, H)
        elif rotation == 270:
            translate_matrix = translate(0, H)

        # Order matters here - the transformation matrices are applied in
        # reverse order. So first we scale to get the points in PDF user space,
        # since all other operations are in that space. Then we rotate and
        # scale to capture page rotation, then finally we translate to account
        # for offset media boxes.
        transform = matrix_multiply(
            mb_translate,
            translate_matrix,
            rotation_matrix,
            scale_matrix,
        )
        return transform

    def _add_annotation(self, annotation):
        """Add the annotation to the PDF document, transforming annotation
        metadata and content stream to PDF user space.
        """
        page = self._pdf.get_page(annotation.page)
        transform = self.get_transform(
            annotation.page,
            self._pdf.get_rotation(annotation.page),
        )
        annotation_obj = annotation.as_pdf_object(transform, page)

        if page.Annots:
            page.Annots.append(annotation_obj)
        else:
            page.Annots = [annotation_obj]

    def write(self, filename=None, overwrite=False):
        if filename is None and not overwrite:
            raise ValueError(
                'Must specify either output filename or overwrite flag'
            )
        if overwrite:
            filename = self._filename

        writer = PdfWriter(
            version=self._pdf.pdf_version,
            compress=self._compress,
        )
        writer.write(fname=filename, trailer=self._pdf._reader)



# -*- coding: utf-8 -*-
from pdf_annotate.annotator import PdfAnnotator
from pdf_annotate.config.appearance import Appearance
from pdf_annotate.config.location import Location
from pdf_annotate.config.metadata import Metadata


__all__ = ['PdfAnnotator', 'Appearance', 'Location', 'Metadata']



# -*- coding: utf-8 -*-
"""
    Points annotations
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    Annotations that are defined by a series of points: Line, Polygon, Polyline

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
from pdf_annotate.annotations.base import Annotation
from pdf_annotate.annotations.base import make_border_dict
from pdf_annotate.config.appearance import set_appearance_state
from pdf_annotate.config.appearance import stroke_or_fill
from pdf_annotate.graphics import Close
from pdf_annotate.graphics import ContentStream
from pdf_annotate.graphics import Line as CSLine
from pdf_annotate.graphics import Move
from pdf_annotate.graphics import Restore
from pdf_annotate.graphics import Save
from pdf_annotate.graphics import Stroke


def flatten_points(points):
    return [v for point in points for v in point]


class PointsAnnotation(Annotation):
    """An abstract annotation that defines its location on the document with
    an array of points.
    """

    def make_rect(self):
        L = self._location
        stroke_width = self._appearance.stroke_width
        p = L.points[0]
        min_x, max_x, min_y, max_y = p[0], p[0], p[1], p[1]
        for x, y in L.points:
            min_x = min(min_x, x)
            max_x = max(max_x, x)
            min_y = min(min_y, y)
            max_y = max(max_y, y)
        return [
            min_x - stroke_width,
            min_y - stroke_width,
            max_x + stroke_width,
            max_y + stroke_width,
        ]

    def base_points_object(self):
        obj = self.make_base_object()
        obj.BS = make_border_dict(self._appearance)
        obj.C = self._appearance.stroke_color
        # TODO line endings, leader lines, captions
        return obj


class Line(PointsAnnotation):
    subtype = 'Line'

    def make_appearance_stream(self):
        A = self._appearance
        points = self._location.points

        stream = ContentStream([Save()])
        set_appearance_state(stream, A)
        stream.add(Move(points[0][0], points[0][1]))
        stream.add(CSLine(points[1][0], points[1][1]))
        stroke_or_fill(stream, A)
        stream.add(Restore())

        return stream

    def add_additional_pdf_object_data(self, obj):
        # TODO line endings, leader lines, captions
        obj.L = flatten_points(self._location.points)


class Polygon(PointsAnnotation):
    subtype = 'Polygon'
    versions = ('1.5', '1.6', '1.7')

    def make_appearance_stream(self):
        A = self._appearance
        points = self._location.points

        stream = ContentStream([Save()])
        set_appearance_state(stream, A)
        stream.add(Move(points[0][0], points[0][1]))
        for x, y in points[1:]:
            stream.add(CSLine(x, y))
        stream.add(Close())
        stroke_or_fill(stream, A)
        stream.add(Restore())

        return stream

    def add_additional_pdf_object_data(self, obj):
        if self._appearance.fill:
            obj.IC = self._appearance.fill
        obj.Vertices = flatten_points(self._location.points)


class Polyline(PointsAnnotation):
    subtype = 'PolyLine'
    versions = ('1.5', '1.6', '1.7')

    def make_appearance_stream(self):
        A = self._appearance
        points = self._location.points

        stream = ContentStream([Save()])
        set_appearance_state(stream, A)
        stream.add(Move(points[0][0], points[0][1]))
        for x, y in points[1:]:
            stream.add(CSLine(x, y))
        # TODO add a 'close' attribute?
        stream.extend([Stroke(), Restore()])

        return stream

    def add_additional_pdf_object_data(self, obj):
        obj.Vertices = flatten_points(self._location.points)


class Ink(PointsAnnotation):
    subtype = 'Ink'

    def make_appearance_stream(self):
        A = self._appearance
        points = self._location.points

        stream = ContentStream([Save()])
        set_appearance_state(stream, A)
        stream.add(Move(points[0][0], points[0][1]))
        # TODO "real" PDF editors do smart smoothing of ink points using
        # interpolated Bezier curves.
        for x, y in points[1:]:
            stream.add(CSLine(x, y))
        stream.extend([Stroke(), Restore()])

        return stream

    def add_additional_pdf_object_data(self, obj):
        obj.InkList = [flatten_points(self._location.points)]



from setuptools import find_packages
from setuptools import setup

setup(
    name='pdf-annotate',
    use_scm_version=True,
    description='Add annotations to PDFs',
    author='Michael Bryant',
    author_email='smart-recordset@plangrid.com',
    url='https://github.com/plangrid/pdf-annotate',
    packages=find_packages('.', exclude=['tests*']),
    include_package_data=True,
    install_requires=[
        'attrs>=18.1.0',  # this could probably be lower, but it's not tested
        'pdfrw>=0.4',
        'pillow>=5.2.0',  # this could probably be lower, but it's not tested'
        'fonttools>=3.44.0'
    ],
    extras_require={
        'tests': [
            'pre-commit',
            'pytest',
            'coverage',
            'flake8',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ]
)
